/* Bas Aarts. .... This program is the master.c .... */

#include <sys/param.h> 
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "pvm3.h"
#include <signal.h>    

#include <time.h>

#define M2S_INIT1	1
#define M2S_INIT2	2
#define M2S_SAVE	3
#define M2S_NOSYNC	4
#define M2S_DATA1	5
#define M2S_DATA2       6

#define S2M_KILL        11
#define S2M_END         12
#define S2M_SYNC        13
#define S2M_OPT		14

#define FALSE           0
#define TRUE           	1

#define INTSIZE         sizeof(int)

#define MAXINT          100000000
#define MAXFILENAMELEN	50
#define ENCODING        PvmDataRaw

typedef char            Bool;
typedef struct {
	  int Cost;
	  int Host;
	}		Pair;

char HostName[MAXHOSTNAMELEN];
struct pvmhostinfo *HostInfo;
int NbrOfHosts;
int NbrOfArchs;
int *ChildTids;
int TasksRunning;
int *NewSol;
char *OutputFile;
int MaxBackNr;
int MaxIt;
int MinIt;
int SyncTime;
int N;
int M;
int J;
int *m;
int *d;

void QuitAllTasks() {
  system("echo reset|pvm");
  pvm_exit();
  exit(0);
}

void Init() {
  if (pvm_mytid() == PvmSysErr) {
    printf("ERROR, pvmd is not responding!\n");
    printf("Make sure that you've constructed a virtual machine!\n");
    pvm_exit();
    exit(0);
  }

  pvm_setopt(PvmRoute, PvmRouteDirect);

  setbuf(stdout,NULL);

  signal(SIGINT, QuitAllTasks);
}

void WriteHelp(char *ProgramName) {
  printf("\nUsage: %s -iInput -oOutput [-MInt1] [-mInt2] [-bInt3] [-sInt] [-tInt5] [-d]\n\n", ProgramName);
  printf(" -iInput   \"Input\" is the instancefile.\n");
  printf(" -oOutput  The output is written to the file \"Output\".\n");
  printf(" -MInt1  Int1 is the maximal number of iterations on a schedule.\n");
  printf(" -mInt2  Int2 is the minimal number of iterations on a schedule.\n");
  printf(" -bInt3  Int3 is the number of intensification schedules.\n");
  printf(" -sInt4  Int4 is the time in seconds between two synchronisations.\n");
  printf(" -tInt5    Terminate the program after Int5 seconds.\n");
  printf(" -d        Save this setting.\n\n");
  pvm_exit();
  exit(0);
}

void CheckInput(int NbrOfArg, char *SetOfArg[]) {
  char *InstanceFile;
  int   i,
    Check   = 0;
  Bool  SaveDef = FALSE;  
  FILE  *fp     = fopen("default.dat", "r");

  if (fp) {
    fscanf(fp, "%d%d%d%d", &MaxBackNr, &MaxIt, &MinIt, &SyncTime);
    fclose(fp);
  }
  else SaveDef = TRUE;

  for (i = 1; i < NbrOfArg; i++) {
    if (*SetOfArg[i] != '-') WriteHelp(*SetOfArg);
    switch (SetOfArg[i][1]) {
      case 'i': InstanceFile = &SetOfArg[i][2]; Check++; break;
      case 'o': OutputFile = &SetOfArg[i][2]; Check++; break;
      case 'M': MaxIt = atoi(&SetOfArg[i][2]); break;
      case 'm': MinIt = atoi(&SetOfArg[i][2]); break;
      case 's': SyncTime = atoi(&SetOfArg[i][2]); break;
      case 'b': MaxBackNr = atoi(&SetOfArg[i][2]); break;
      case 'd': SaveDef = TRUE; break;
      case 't':
		signal(SIGALRM, QuitAllTasks);
		alarm(atoi(&SetOfArg[i][2]));
      default : WriteHelp(*SetOfArg);
    }
  }

  if (Check < 2) WriteHelp(*SetOfArg);

  if (SaveDef) {
    fp = fopen("default.dat", "w");
    fprintf(fp, "%d\n%d\n%d\n%d\n", MaxBackNr, MaxIt, MinIt, SyncTime);
    fclose(fp);
  }

  if (!(fp = fopen(InstanceFile, "r"))) {
    printf("ERROR, can't open file %s!\n", InstanceFile);
    pvm_exit();
    exit(0);
  }
  fscanf(fp,"%d %d", &J, &M);
  m = (int*) calloc(N = J*M, INTSIZE);
  d = (int*) calloc(N, INTSIZE);
  for(i =0; i < N; i++) fscanf(fp, "%d %d", &m[i], &d[i]);
  fclose(fp);
}

void PrintMachineInfo(char* ProgramName) {
  if (pvm_config(&NbrOfHosts, &NbrOfArchs, &HostInfo) == PvmSysErr) {
    printf("ERROR, pvmd is not responding!\n");
    printf("Make sure that you've constructed a virtual machine!\n");
    pvm_exit();
    exit(0);
  }

  gethostname(HostName, MAXHOSTNAMELEN);

  printf("Program %s is running on host %s.\n", ProgramName, HostName);

  switch (NbrOfHosts) {
    case 1 : printf("There are 0 slaves available.\n");
	     pvm_exit();
	     exit(0);
    case 2 : printf("There is 1 slave available.\n"); break;
    default: printf("There are %d slaves available.\n", NbrOfHosts-1);
  }
}

void StartSlaves() {
  int i;

  ChildTids = (int*) calloc(INTSIZE, NbrOfHosts-1);
  printf("Spawning slaves.");
  for(i = TasksRunning = 0; i < NbrOfHosts; i++)
    if (strcmp(HostName, HostInfo[i].hi_name)) {

      int j =
	pvm_spawn("ts2",                  
		  (char**) 0,             
		  PvmTaskHost,            
		  HostInfo[i].hi_name,    
		  1,                      
		  &ChildTids[TasksRunning]
		  );
      if (j < 0) {
	printf("\nERROR, can't spawn task on host %s!\n", HostInfo[i].hi_name);
	QuitAllTasks();
      }
      else if (j > 0) TasksRunning++;
    };
  printf(" Done!\n");

  if (TasksRunning == 0) {
    printf("ERROR, no tasks were started!\n");
    pvm_exit();
    exit(0);
  }
  if (TasksRunning != NbrOfHosts-1) {
    printf("WARNING, only %d of the %d tasks %s started!\n",
	   TasksRunning,
	   NbrOfHosts-1,
	   TasksRunning == 1 ? "was" : "were"
	  );
    printf("Continue with execution? [y/n] : ");
    if ((char) getchar() != 'y') {
      printf("\nWARNING, program terminated!\n");
      QuitAllTasks();
    }
  }

  pvm_notify(PvmTaskExit, S2M_KILL, TasksRunning, ChildTids);
}

void BroadCastInfo() {
  int i;

  printf("Broadcasting information to slaves.");

  for (i = 0; i < TasksRunning; i++) {
    pvm_initsend(ENCODING);
    pvm_pkint(&i, 1, 1);
    pvm_send(ChildTids[i], M2S_INIT1);
  }

  pvm_initsend(ENCODING);
  pvm_pkint(&M, 1, 1);
  pvm_pkint(&J, 1, 1);
  pvm_pkint(m, N, 1);
  pvm_pkint(d, N, 1);
  pvm_pkint(&MaxBackNr, 1, 1);
  pvm_pkint(&MaxIt, 1, 1);
  pvm_pkint(&MinIt, 1, 1);
  pvm_pkint(&SyncTime, 1, 1);

  pvm_mcast(ChildTids, TasksRunning, M2S_INIT2);
  printf(" Done!\n");
}

int SyncSort(const void *a, const void *b) {
  return ((Pair*)a)->Cost - ((Pair*)b)->Cost;
}

int SyncSort1(const void *a, const void *b) {
  return NewSol[*(int*) b] - NewSol[*(int*) a];
}

void ReceiveDataFromSlaves() {
  int ReadyToSync = 0;
  int NbrOfSol = 0;

  Pair *CurSol = (Pair*) calloc(sizeof(Pair), TasksRunning*MaxBackNr);

  int NbrOfNewSol;

  int *Waste = (int*) calloc(INTSIZE, TasksRunning);

  int *NewSolOrd = (int*) calloc(INTSIZE, TasksRunning);

  int Tid;

  int Tag;

  int i, j, k;

  int Optimum = MAXINT;

  int OptTid;

  int NbrOfTasksStarted = TasksRunning;

  NewSol = (int*) calloc(INTSIZE, TasksRunning);
  for (i = 0; i < TasksRunning; i++) 
    NewSolOrd[i] = i;
  
  while (TasksRunning) {
    pvm_bufinfo(pvm_recv(-1, -1), (int*) 0, &Tag, &Tid);
    switch (Tag) {
    case S2M_KILL:
      pvm_upkint(&i, 1, 1);
      for (j = 0, i = pvm_tidtohost(i); HostInfo[j].hi_tid != i; j++);
      printf("ERROR, failure on host %s!\n", HostInfo[j].hi_name);
      QuitAllTasks();

    case S2M_END:
      pvm_upkint(&i, 1, 1);
      pvm_upkint(&j, 1, 1);
      if (j < Optimum) {
	Optimum = j;
	OptTid = Tid;
	printf("%d\n", Optimum);
      }
      NewSol[i] = -1;
      TasksRunning--;
      if (TasksRunning == 0) break;

    case S2M_SYNC:
      if (Tag == S2M_SYNC) {
	pvm_upkint(&k, 1, 1);
	pvm_upkint(&j, 1, 1);
	for (;j ;j--){
	  pvm_upkint(&CurSol[NbrOfSol].Cost, 1, 1);
	  CurSol[NbrOfSol++].Host = k;
	}
	NewSol[k] = Waste[k] = 0;
	ReadyToSync++;
      }

      if (ReadyToSync == TasksRunning) {
	if (TasksRunning > 1) {

	  static float Epsilon = 0.075;
	  float Maximum;

	  qsort(CurSol, NbrOfSol, sizeof(Pair), SyncSort);

	  if (CurSol[0].Cost < Optimum) {
	    Optimum = CurSol[0].Cost;
	    OptTid = ChildTids[CurSol[0].Host];
	    printf("%d\n", Optimum);
	  }

	  Maximum = ((float) CurSol[0].Cost)*(1.0 + Epsilon);
	  for (i = 0; i < NbrOfSol &&
		 ((float) CurSol[i].Cost < Maximum || i%TasksRunning); i++)
	    NewSol[CurSol[i].Host]++;
	  for (NbrOfNewSol = i; i < NbrOfSol; i++) 
	    Waste[CurSol[i].Host]++;
	  qsort(NewSolOrd, NbrOfTasksStarted, INTSIZE, SyncSort1);
	  
	  int i1 = NbrOfNewSol/TasksRunning, 
	    i2 =  NbrOfNewSol%Tasks = Running;

	  for (i = 0; i < TasksRunning; i++) {
	    j = NewSol[NewSolOrd[i]] - i1 - (i < i2 ? 1 : 0);
	    pvm_initsend(ENCODING);
	    pvm_pkint(&Waste[NewSolOrd[i]], 1, 1);
	    pvm_pkint(&j, 1, 1);
	    pvm_send(ChildTids[NewSolOrd[i]], M2S_DATA1);
	  }

	  i = 0; j = TasksRunning - 1;
	  while (i != j)
	    {
              if (NewSol[NewSolOrd[i]] == i1 + (i < i2 ? 1 : 0)) i++;
              else
              if (NewSol[NewSolOrd[j]] == i1 + (j < i2 ? 1 : 0)) j--;
              else {
                pvm_initsend(ENCODING);
                pvm_pkint(&ChildTids[NewSolOrd[j]], 1, 1);
                pvm_send(ChildTids[NewSolOrd[i]], M2S_DATA2);
                NewSol[NewSolOrd[i]]--;
                NewSol[NewSolOrd[j]]++;
              }
            }

	  if (Epsilon > .005) Epsilon /= 1.5;
            NbrOfSol = 0;
	}
	else { 
	  pvm_initsend(ENCODING);
	  pvm_send(ChildTids[k], M2S_NOSYNC);
	}
	ReadyToSync = 0;
      }
      break;

    case S2M_OPT:
      pvm_upkint(&Optimum, 1, 1);
      OptTid = Tid;
      TasksRunning = 0;
    }
  }
  printf("Best solution has cost %d.\n", Optimum);
  pvm_initsend(ENCODING);
  pvm_pkstr(OutputFile);
  pvm_send(OptTid, M2S_SAVE);
  pvm_bufinfo(pvm_recv(-1, -1), (int*) 0, &Tag, &Tid);
  if (Tag == S2M_KILL) {
    pvm_upkint(&i, 1, 1);
    for (j = 0, i = pvm_tidtohost(i); HostInfo[j].hi_tid != i; j++);
    printf("ERROR, failure on host %s!\n", HostInfo[j].hi_name);
    QuitAllTasks();
  }
  /* else OK */
}

int main(int NbrOfArg, char *SetOfArg[]) {
  int Start1, Start2;
  Init();
  CheckInput(NbrOfArg, SetOfArg);
  PrintMachineInfo(*SetOfArg);
  Start1 = time(NULL);
  StartSlaves();
  BroadCastInfo();
  Start2 = time(NULL);
  ReceiveDataFromSlaves();
  printf("Program terminated in %d seconds.\n", time(NULL) - Start1);
  printf("Initialization took %d seconds.\n", Start2 - Start1); =
  QuitAllTasks();
}

